# 3D People Counting-ISK V1884R
# ver:0.0.1
# 2023/01/03
# firware: V1884R
# parsing People Counting 3D fusion  
# hardware:(Batman-201)ISK IWR6843
# 
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
#===========================================
# output: v1010,v1011,v1012,v1020,v1021 Raw data
# v0.0.1 : 2023/01/05 release
#          (1)Output list data
# v0.0.2 : 2023/04/07 add try...except
# v0.0.3 : 2023/07/12 add recording_v2, readFile_v2 ,getRecordData_v2
# v0.0.4 : 2024/03/13 fix unpack requires a buffer of 8 bytes 
#

import serial
import time
import struct
import numpy as np
import pandas as pd


class header:
	version = 0
	totalPackLen = 0
	platform = 0
	frameNumber = 0
	timeCpuCycles = 0
	numDetectedObj = 0
	numTLVs = 0
	subframeNumber = 0

class unitS:
	elevationUnit:float = 0.0
	azimuthUnit : float = 0.0
	dopplerUnit : float = 0.0
	rangeUnit :   float = 0.0
	snrUnit :     float = 0.0

class pc3_v1884R:
	gdata = b''
	playback = False
	#magicWord =  [b'\x02',b'\x01',b'\x04',b'\x03',b'\x06',b'\x05',b'\x08',b'\x07',b'\0x99']
	magicWord =  [b'\x01',b'\x02',b'\x03',b'\x04',b'\x05',b'\x06',b'\x07',b'\x08',b'\0x99']
	port = ""
	hdr = header
	u = unitS
	frameNumber = 0
	# provide csv file dataframe
	# real-time 
	v1020_names_rt = ['fN','sx','sy','sz','ran','elv','azi','dop','snr','fn']
	v1010_col_names_rt = ['fN','type','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','ec0','ec1','ec2','ec3','ec4','ec5','ec6','ec7','ec8','ec9','ec10','ec11','ec12','ec13','ec14','ec15','g','confi','tid']                    
	 
	vxx_col_names = ['fN','type','i0','i1','i2','i3','i4', 'i5','i6' ,'i7','i8','i9']  
	# read from file for trace point clouds
	fileName = ''
	v1020_col_names = ['time','fN','type','sx','sy','sz','ran','elv','azi', 'dop', 'snr','fn']
	
	
	v1010_col_names = ['time','fN','type','tid','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','ec0','ec1','ec2','ec3','ec4','ec5','ec6','ec7','ec8','ec9','ec10','ec11','ec12','ec13','ec14','ec15','g','confi']
	 
	v1020simo = []
	
	sim_startFN = 0
	sim_stopFN  = 0 
	JB_t0 = 0
	version_keep = '3050004'
	
	# add for interal use
	tlvLength = 0
	numOfPoints = 0
	# for debug use 
	dbg = False #Packet unpacket Check: True show message 
	sm  = False #Observed StateMachine: True Show message
	plen = 16 
	
	def __init__(self,port, azi_degree = None, bufSize = None):
		
		port.reset_input_buffer()
		port.ReadBufferSize = bufSize
		self.degree = azi_degree
		self.port = port
		self.azi_offset = 0.0 if azi_degree == None else azi_degree * np.pi/180.0
		print("(jb)People Counting 3D initial")
		print("(jb)version:v0.1.0")
		print("(jb)For Hardware:Batman-201(ISK)/AOP(501/601)")
		print("(jb)Firmware: PC3")
		print("(jb)UART Baud Rate:921600")
		print("Output: (dck,v1010,v1011,v1012,v1020,v1021) (RAW data)\n##########################################\n\n\n")
		
	def useDebug(self,ft):
		self.dbg = ft
		
	def stateMachine(self,ft):
		self.sm = ft
		
	def getHeader(self):
		return self.hdr
		
	def headerShow(self):
		print("***header***********") 
		print("Version:     \t%x "%(self.hdr.version))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
		print("Platform:    \t%X "%(self.hdr.platform))
		print("PID(frame#): \t%d "%(self.hdr.frameNumber))
		print("CPU cycles Time:\t{:d} us".format(self.hdr.timeCpuCycles))
		print("Number of Detected Object: \t%d "%(self.hdr.numDetectedObj))
		print("numTLVs:     \t%d "%(self.hdr.numTLVs))
		print("subframe#  : \t%d "%(self.hdr.subframeNumber))
		print("***End Of Header***") 
		
#################################################################
	JB_syncFlag = 0
	def jb_getUartBuf(self,disp = None):
		idx = 0
		buf = b""
		self.gdata = b''
		while True:
			if self.JB_syncFlag == 0:
				ch = self.port.read() # here ch type is byte
				if ch == self.magicWord[idx]: # 0..7
					buf += ch
					idx += 1 # 1..8
					if idx == 8:
						idx = 8
						self.JB_syncFlag = 1
						self.JB_t0 = time.time()
			else:
				self.JB_syncFlag = 0 # magic word ch mismatch loop back again
				idx = 0 # init
				buf = b""
				self.gdata = b''
			

			if self.JB_syncFlag == 1:
				# read next 16 bytes for v, l, p, f
				for i in range(32):
					ch = self.port.read()
					buf += ch
					self.gdata = buf

				if len(buf) == (32+8): # magic+header
					try:
						(self.hdr.version,self.hdr.totalPackLen,self.hdr.platform,self.hdr.frameNumber,
						self.hdr.timeCpuCycles,self.hdr.numDetectedObj,self.hdr.numTLVs,self.hdr.subFrameNumber) = struct.unpack('8I',buf[8:40]) 
						self.frameNumber = self.hdr.frameNumber
						verString = "{:X}".format(self.hdr.version)
						totalLen = self.hdr.totalPackLen
					except:
						if self.dbg == True:
							print("unpack except:")
						return False
						
					if disp:
						self.headerShow()
					
					if self.playback == True:
						return False
					
					if verString != self.version_keep or self.hdr.totalPackLen > 100000:
						print("ver: {:}   version_keep:{:}".format(verString,self.version_keep ))
						return False

				else:
					self.JB_syncFlag = 0 # len mismatch loop back again
					idx = 0
					buf = b''
					self.gdata = b''

			if self.JB_syncFlag == 1:
				r = self.port.read(totalLen - 40) # read rest of bytes
				#xx = self.port.inWaiting()
				#print("is Wating= {:}  :len:{:}".format(0,len(r)))
				#buf += r
				self.gdata += r
				dt = time.time()- self.JB_t0
				#self.port.flushInput()
				self.port.reset_input_buffer()
				 
				#print(f"==== gdata_len: {len(self.gdata)}" )
				if totalLen  == len(self.gdata):
					#print(f"====eq gdata_len: {len(self.gdata)}" )
					return True
				else:
					self.JB_syncFlag = 0 # len mismatch loop back again
					idx = 0
					buf = b""
					self.gdata = b''
					
	def check_existence(self, typeList, data):
		return data in typeList
		
#################################################################
#
# TLV: Type-Length-Value
# read TLV data
# input:
#     disp: True:print message
#			False: hide printing message
#     df: 
# output:(return parameter)
#  (dck,v1010,v1011,v1012,v1020,v1021) = radar.tlvRead(False)
#  dck: True: Data available    False: Data not available
#
#	v1010 =([]) #target
#	v1011 =([]) #target index
#	v1012 =([]) #target height
#	v1020 =([]) #point  cloud
#	v1021 =([]) #Prescence Indication
#
#================================================================
	 
	def tlvRead(self,disp):
		global JB_tOld 
		 
		#print("---tlvRead---")
		#ds = dos
		#typeList = ['1010','1011','1012','1020','1021']
		typeList = [1020,1011,1012,1010,1021]
		idx = 0
		lstate = 'idle'
		sbuf = b''
		lenCount = 0
		 
		dataBytes = 0
		numOfPoints = 0
		tlvCount = 0
		 
		v1010 =([]) #target
		v1011 =([]) #target index
		v1012 =([]) #target height
		v1020 =([]) #point cloud
		v1021 =([]) #Prescence Indication
		
		ch = b'\x00' # init
		idxTo = 0
		uartBuf = b''
		 
		while True:
			###############################
			# on {idle} getUartBuf() else parsing from uartBuf[]
			if lstate == 'idle':
				chk = self.jb_getUartBuf(disp=disp)
				if chk == True:
					uartBuf = self.gdata
					tlvCount = self.hdr.numTLVs
				else:
					#print("===========================tlvRead=========================buf empty" )
					return (False,v1010,v1011,v1012,v1020,v1021)
					
				lstate = 'TL'
				sbuf = b"" # init
				idxFrom = 0
				idxTo = 40  #header 32 + 8(magic word) = 40
				
				if self.sm:
					print("===============================")
					print(f"JB> (idle)->(TL):    numTLVs(tlvCount):{tlvCount}")
				
				
			elif lstate == 'TL': #TLV Header type/length
				idxFrom = idxTo 
				idxTo += 8
				try:
					if len(uartBuf[idxFrom : idxTo]) == 8: 
						#print(f'========uartBuf length: {len(uartBuf[idxFrom : idxTo])}   :  uartBuf:{len(uartBuf)} gdata:{len(self.gdata)} index[From, to]:{idxFrom}: {idxTo}======================')
						ttype, self.tlvLength = struct.unpack('2I', uartBuf[idxFrom : idxTo])
						ck = self.check_existence(typeList,ttype)
						#print(f'========ttype: {str(ttype)} self.tlvLength: {self.tlvLength}  check:{ck} '  )
						if ck == False:
							return (False,v1010,v1011,v1012,v1020,v1021)
					else:
						#print("************************error  ttype, self.tlvLength = struct.unpack('2I' **********************************" )
						return (False,v1010,v1011,v1012,v1020,v1021)
						
				except:
					if self.dbg == True:
						print("Header unpack except")
					return (False,v1010,v1011,v1012,v1020,v1021)
						
				if self.sm:
					print(f"JB> (TL)->({ttype}) indexFrom:idxTo= {idxFrom}:{idxTo}  type:{ttype} tlvLength:{self.tlvLength}")
				lstate = str(ttype)
			
			elif lstate == '1020': # count = Total Lentgh - 8   # eadrs: 8 bytes
				
				(points,dataBytes) = self.points_cal(dataBytes = 8, offset = 20) 
				datalen = self.tlvLength - 20 # tlv(8 bytes) units(4x5=20 bytes)
				 
				
				#print("JB> (1020) points: {:}    datalen:{:}  idxTo:{:}".format(points,datalen,idxTo))
				#(0) unit unpack
				idxFrom = idxTo
				idxTo += 20
				try:
					self.u.elevationUnit,self.u.azimuthUnit,self.u.dopplerUnit,self.u.rangeUnit,self.u.snrUnit = struct.unpack('5f', uartBuf[idxFrom : idxTo])
				except:
					if self.dbg == True:
						print("Header unpack except")
					return (False,v1010,v1011,v1012,v1020,v1021)
				#print("JB> (1020-unit)  ==> elv:{:.4f} azimuth:{:.4f} doppler:{:.4f} range:{:.4f} snr:{:.4f}".format(self.u.elevationUnit,self.u.azimuthUnit,self.u.dopplerUnit,self.u.rangeUnit,self.u.snrUnit))
				
				#(1) reads unpack
				idxFrom = idxTo
				idxTo += datalen
				
				#sbuf = uartBuf[idxFrom : idxTo]
				for i in range(points):
					try: 
						(e,a,d,r,s) = struct.unpack('2bh2H', uartBuf[idxFrom + i*dataBytes : idxFrom + (i + 1)*dataBytes]) #8bytes
						elv = e * self.u.elevationUnit
						azi = a * self.u.azimuthUnit  + self.azi_offset
						dop = d * self.u.dopplerUnit
						ran = r * self.u.rangeUnit
						snr = s * self.u.snrUnit
						sz  = ran * np.sin(elv)
						sx  = ran * np.cos(elv) * np.sin(azi)
						sy  = ran * np.cos(elv) * np.cos(azi) 
						
						#print("({:}:{:4d})(idx:({:4d}) elv:{:.4f} azimuth:{:.4f} doppler:{:.4f} range:{:.4f} snr:{:.4f}".format(points,lenCount,i,elv,azi,dop,ran,snr))
						v1020.append((sx,sy,sz,ran,elv,azi,dop,snr,self.hdr.frameNumber)) #xyzreadsf
						
					except:
						if self.dbg == True:
							print("(1020)Improper Type 1020 Value structure found: ")
						return (False,v1010,v1011,v1012,v1020,v1021)
				 
				(chk,lstate,tlvCount) = self.idle_tl_check(tlvCount = tlvCount,state= lstate)
				if chk:
					return (True,v1010,v1011,v1012,v1020,v1021)
				
						
			elif lstate == '1010': # Target list
				(points,dataBytes) = self.points_cal(dataBytes=112,offset=0) #112 = 4 + 9 * 4 + 18 *4 
				idxFrom = idxTo
				idxTo += self.tlvLength #    datalen
				#print("JB> (1010) points: {:}    datalen:{:}  idxTo:{:}".format(points,self.tlvLength,idxTo))
				#print("JB> (1010) idxFrom: {:}  idxTo:{:}".format(idxFrom,idxTo))
				
				for i in range(points):
					try:
						 
						(tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,
						ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi) = struct.unpack('I27f', uartBuf[idxFrom + i*dataBytes : idxFrom + (i + 1)*dataBytes])
						
						#posX, posY = rotate_matrix(pX, pY, self.azi_offset  , x_shift=0, y_shift=0, units="RADIAN")  
						#print(f'JB>(1010) point({i}) tid:{tid} posX:{posX} posY:{posY} posZ:{posZ} ')
						v1010.append((tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi))
						
					except:
						if self.dbg == True:
							print("(7)Improper Type 1010 Value structure found: ")
						return (False,v1010,v1011,v1012,v1020,v1021)
				
				(chk,lstate,tlvCount) = self.idle_tl_check(tlvCount = tlvCount,state= lstate)
				if chk:
					return (True,v1010,v1011,v1012,v1020,v1021)
				
			
			elif lstate == '1011': #Target Index
				(points,dataBytes) = self.points_cal(dataBytes=1,offset=0)
				idxFrom = idxTo
				idxTo += self.tlvLength
				#print(f'(v1011) type:{ttype} length:{self.tlvLength} ')
				for i in range(points): #v2
					try:
						tidx = struct.unpack('B', uartBuf[idxFrom + i*dataBytes : idxFrom + (i + 1)*dataBytes])
						#print(f'JB>(1011) point({i}) tidx:{tidx}')
						v1011.append(tidx)
					except:
						if self.dbg == True:
							print("(8)Improper Type 1011 Value structure found: ")
						
						return (False,v1010,v1011,v1012,v1020,v1021)
				 
				(chk,lstate,tlvCount) = self.idle_tl_check(tlvCount = tlvCount,state= lstate)
				if chk:
					return (True,v1010,v1011,v1012,v1020,v1021)
				
			
			elif lstate == '1012': #Target Height
				(points,dataBytes) = self.points_cal(dataBytes=12,offset=0)
				
				idxFrom = idxTo
				idxTo += self.tlvLength
				#print(f'(v1012) type:{ttype} length:{self.tlvLength} ')
				for i in range(points): # v2
					try:
						(tid,maxZ,minZ) = struct.unpack('I2f', uartBuf[idxFrom + i*dataBytes : idxFrom + (i + 1)*dataBytes])
						#print(f'JB>(1012) point({i}) tid:{tid} minZ:{minZ} maxZ:{maxZ}  ')
						v1012.append((tid,maxZ,minZ))
					except:
						if self.dbg == True:
							print("(9)Improper Type 1012 Value structure found: ")
						return (False,v1010,v1011,v1012,v1020,v1021)
				 
				(chk,lstate,tlvCount) = self.idle_tl_check(tlvCount = tlvCount,state= lstate)
				if chk:
					return (True,v1010,v1011,v1012,v1020,v1021)
				
				 
					
			elif lstate == '1021': #Prescence Indication
				(points,dataBytes) = self.points_cal(dataBytes=4,offset=0)
				idxFrom = idxTo
				idxTo += self.tlvLength
			
				#print(f'(v1021) type:{ttype} length:{self.tlvLength} ')
				for i in range(points):
					try:
						pri = struct.unpack('I', uartBuf[idxFrom + i*dataBytes : idxFrom + (i + 1)*dataBytes])
						v1012.append(pri)
					except:
						if self.dbg == True:
							print("(10)Improper Type 1021 Value structure found: ")
						return (False,v1010,v1011,v1012,v1020,v1021)
				
				(chk,lstate,tlvCount) = self.idle_tl_check(tlvCount = tlvCount,state= lstate)
				if chk:
					return (True,v1010,v1011,v1012,v1020,v1021)
				
			else:
				lstat = 'idle'
				#print("===============state error====================")
				return (False,v1010,v1011,v1012,v1020,v1021)
					
	def points_cal(self,dataBytes= None,offset = None): 
		datalen = self.tlvLength - offset
		points = int(datalen/dataBytes)
		return (points,dataBytes)
		
	def idle_tl_check(self,tlvCount = None,state= None,points = None):
		chk = False
		tlvCount -= 1
		if tlvCount == 0:
			if self.sm == True:
				print(f"JB> ({state})->(idle)   tlvCount:{tlvCount}")
			state = 'idle'
			chk = True
			
		else:
			if self.sm == True:
				print(f"JB> ({state})->(TL)  tlvCount:{tlvCount}")
			state = 'TL'
		
		return (chk,state,tlvCount)
		

	def rotate_matrix(self,x, y, angle, x_shift=0, y_shift=0, units="DEGREES"):
		# Shift to origin (0,0)
		x = x - x_shift
		y = y - y_shift
		
		# Convert degrees to radians
		if units == "DEGREES":
			angle = np.radians(angle)

		# Rotation matrix multiplication to get rotated x & y
		xr = (x * np.cos(angle)) - (y * np.sin(angle)) + x_shift
		yr = (x * np.sin(angle)) + (y * np.cos(angle)) + y_shift
		return xr, yr



		
	def getRecordData(self,frameNum):
		s_fn = frameNum +  self.sim_startFN
		print(" s_fn:({:})  = frameNum:{:} + sim_startFN: {:}      ".format(s_fn,frameNum,self.sim_startFN))
		v20d = self.v1020simo[self.v1020simo['fn'] == s_fn]
		 
		print(v20d)
		chk = 0
		if v20d.count != 0:
			chk = 1
		return (chk,v20d)
		
	###############################################################################################
	#
	#V2  recording func:   recording_v2 
	#    read      func:   readFile_v2
	#
	def recording_v2(self,writer= None,v1010=None, v1011= None,v1012= None, v1020 = None, v1021 = None,fn = None):
		if writer is not None:
			#ts = datetime.now()
			
			if v1010 is not None: #ok
				 
				
				if len(v1010) > 0: 
					v1010l = v1010
					for i in v1010l: 
						item = [fn,'v1010'] + list(i[0:10])
						writer.writerow(item)
			if v1011 is not None: #ok
				if len(v1011) > 0:
					it =[]
					for i in v1011: 
						it.append(i[0])
					item = [fn,'v1011'] + [it]
					#writer.writerow(item)
					
			if v1012 is not None:
				if len(v1012) > 0:
					v1012l = v1012
					for i in v1012l: 
						item = [fn,'v1012'] + list(i)
						writer.writerow(item)
			
			if v1020 is not None:  #ok
				if len(v1020) > 0:
					v1020l = v1020
					for i in v1020l: 
						item = [fn,'v1020'] + list(i)
						writer.writerow(item)
						
			if v1021 is not None:
				if len(v1021) > 0:
					v1021l = v1021
					for i in v1021l: 
						item = [fn,'v1021'] + list(i)
						writer.writerow(item)
		
	def readFile_v2(self,fileName):
		#fileName = "pc32021-03-19-10-02-17.csv"  
		#df = pd.read_csv(fileName, error_bad_lines=False, warn_bad_lines=False) 
		self.fileName = fileName 
		df = pd.read_csv(self.fileName)
		
		
		self.v1010simo = df[df['type'] == 'v1010'].rename(columns={'i0': 'tid', 'i1': 'posX', 'i2': 'posY', 'i3': 'posZ', 'i4': 'velX', 'i5': 'velY', 'i6': 'velZ', 'i7': 'accX', 'i8': 'accY', 'i9': 'accZ'})
		#print("------------------- df --------------------shape:{:}".format(df.shape))
		print(df.info())
		#print(df.info(memory_usage="deep")) 
		 
		# (sx,sy,sz,ran,elv,azi,dop,snr,self.hdr.frameNumber)
		self.v1020simo = df[df['type'] == 'v1020'].drop(['i9'], axis=1).rename(columns={'i0': 'X', 'i1': 'Y', 'i2': 'Z', 'i3': 'range', 'i4': 'elv', 'i5': 'azi','i6': 'dop', 'i7': 'snr','i8':'fn'})
		
		
		self.sim_startFN = int(df['fN'].values[1])
		self.sim_stopFN  = int(df['fN'].values[-1])
		
		#print(f'self.sim_startFN  = {self.sim_startFN} self.sim_stopFN = {self.sim_stopFN} ')
		
		return (self.sim_startFN,self.sim_stopFN,self.v1020simo,self.v1010simo)
	
	
	
		
	def getRecordData_v2(self,frameNum):
		s_fn = frameNum +  self.sim_startFN
		print(" s_fn:({:})  = frameNum:{:} + sim_startFN: {:}      ".format(s_fn,frameNum,self.sim_startFN))
		v20d = self.v1020simo[self.v1020simo['fN'] == s_fn]
		v10d = self.v1010simo[self.v1010simo['fN'] == s_fn] 
		
		chk = 0
		if v20d.count != 0:
			chk = 1
		return (chk,v20d,v10d)
	
	###############################################################################################
			
		
	def readFile(self,fileName):
		#fileName = "pc32021-03-19-10-02-17.csv"  
		#df = pd.read_csv(fileName, error_bad_lines=False, warn_bad_lines=False) 
		self.fileName = fileName 
		#          ['time','fN','type','elv','azimuth','range' ,'doppler','snr','sx', 'sy', 'sz']
		df = pd.read_csv(self.fileName, names = self.v1020_col_names,skiprows = [0]) 
		df.dropna()
		#print("------------------- df --------------------shape:{:}".format(df.shape))
		
		print(df.info())
		print(df.info(memory_usage="deep")) 
		
		v1020simori = df[(df.type == 'v1020')]
		print("=============v1020simori=================")
		print(v1020simori)
		print("=============v1020simori=  xxx================")
		self.sim_startFN = df['fN'].values[0]
		self.sim_stopFN  = df['fN'].values[-1]
		
		print(f"sim_start: {self.sim_startFN} sim_stop: {self.sim_stopFN}")
		
		#print("-------------------v1020simo------------:{:}".format(v1020simori.shape))
		
		self.v1020simo = v1020simori.loc[:,['sx', 'sy', 'sz','ran' ,'elv','azi','dop','snr','fn']] # in this case
		#self.v1020simo['fn'] = self.v1020simo['fn'].astype(int, errors = 'raise') 
		
		#print(self.v1020simo)
		#print("-------------------v1020simo--------  end  ---------")
		
		 
		df10 = pd.read_csv(self.fileName, names = self.v1010_col_names, skiprows = [0])  
		
		#------------- v1010 sim ---------------
		v1010simc = df10[df10['type'] == 'v1010']
		self.v1010simo  = v1010simc.loc[:,['fN','type','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','tid']]
		self.v1010simo['posX'] = self.v1010simo['posX'].astype(float, errors = 'raise') 
		
		
		#------------- v8 sim ---------------
		v8simc = df[df['type'] == 'v8']
		self.v8simo  = v8simc.loc[:,['fN','type','elv']]
		self.v8simo.columns = ['fN','type','targetID']
		
		#print(self.v8simo)
		return (self.v1020simo,self.v1010simo) #,self.v8simo)
		 
		 







# People Counting 3D + VSD  V2 
# 2022/09/12: beta
# v0.0.1
# v0.0.2 change dictionary: swap 'br' & 'hr'
# parsing People Counting 3D + VSD data use AOP
# hardware: BM-501 AOP
#     
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
#===========================================
# output: v1020,v1010,v1011,v1021 & v1040 raw data
# 

import serial
import time
import struct
import pandas as pd
import numpy as np

class header:
	version = 0
	totalPackLen = 0
	platform = 0
	frameNumber = 0
	timeCpuCycles = 0
	numDetectedObj = 0
	numTLVs = 0
	subframeNumber = 0


class vsInfo:
	tid 				 = int(0)
	rangeBin 			 = int(0)
	breathingDeviation   = float(0.0)
	heart_rate 			 = float(0.0)
	breathing_rate 		 = float(0.0)
	vitalSign_heart_buf  = [float(0.0)] * 15
	vitalSign_breath_buf = [float(0.0)] * 15
	

class Pc3_VSD_v2:
	MAGICWORD =  [b'\x01',b'\x02',b'\x03',b'\x04',b'\x05',b'\x06',b'\x07',b'\x08',b'\0x99']
	LEN_MAGIC_HEADER = 40
	port = ""
	hdr = header
	
	vs = vsInfo()  #vsos()
	playback = 0 
	sim_startFN = 0
	sim_stopFN  = 0 
	
	# add for interal use
	tlvLength = 0
	numOfPoints = 0
	# for debug use 
	dbg = False #Packet unpacket Check: True show message 
	sm  = False #Observed StateMachine: True Show message
	plen = 16 
	vitalSign_len = (8 * 11 + 2 * 4)
	#vsdic = {}
	vsdic= {'tid':0,'rangeBin' : 0,
					'breathDev': 0, 'br' : 0  ,'hr':0,
					'hrBuf': [], 'brBuf':[]
					}
	def __init__(self,port,azi_degree = None, bufSize = None):
		self.port = port
		self.port.reset_input_buffer()
		self.port.ReadBufferSize = bufSize
		self.degree = azi_degree
		self.azi_offset = 0.0 if azi_degree == None else azi_degree * np.pi/180.0
		
		print("\n\n\n(jb)People Counting 3D + VSD V2 initial")
		print("(jb)vsersion:v0.1.0")
		print("(jb)For Hardware:BM-501 (AOP)")
		print("(jb)Hardware: BM501-AOP")
		print("(jb)Firmware:")
		print("(jb)UART Baud Rate:921600")
		print("(jb)Output: v1020,v1010,v1011,v1021 & v1040 data:(RAW)\n\n\n\n")
		 
		
	def useDebug(self,ft):
		self.dbg = ft
		
	def stateMachine(self,ft):
		self.sm = ft
		
	def getHeader(self):
		return self.hdr
		
	def headerShow(self):
		print("******* Header ********") 
		print("Version:     \t%x "%(self.hdr.version))
		print("Platform:    \t%X "%(self.hdr.platform))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
		print("PID(frame#): \t%d "%(self.hdr.frameNumber))
		print("Time in CPU Cycles:\t{:d} us".format(self.hdr.timeCpuCycles))
		print("numTLVs:     \t%d "%(self.hdr.numTLVs))
		print("subframe#  : \t%d "%(self.hdr.subframeNumber))
		print("***End Of Header***") 
			


#
# TLV: Type-Length-Value
# read TLV data
# input:
#     disp: True:print message
#			False: hide printing message
#     df: 
# output:(return parameter)
# (pass_fail,v1020,v1010,v1011,v1021,v1040)
#  pass_fail: True: Data available    False: Data not available
#  v1020: point cloud infomation
#  v1010: Target Object information
#  v1011: Target Index information
#  v1040: Vital Sign information
#  v10421: Presence Indication information
#  
#  v1020,v1010,v1011,v1021,v1040)
	JB_syncFlag = 0
	def jb_getUartBuf(self):
		idx = 0
		buf = b''
		while True:
			#(0) magic word
			if self.JB_syncFlag == 0:
				ch = self.port.read() # here ch type is byte
				if ch == self.MAGICWORD[idx]: # 0..7
					buf += ch
					idx += 1 # 1..8
					if idx == 8:
						idx = 8
						self.JB_syncFlag = 1
						self.JB_t0 = time.time()
			else:
				self.JB_syncFlag = 0 # magic word ch mismatch loop back again
				idx = 0 # init
				buf = b''
			
			#(1) header portion
			if self.JB_syncFlag == 1:
				for i in range(32):
					ch = self.port.read()
					buf += ch
				if len(buf) == self.LEN_MAGIC_HEADER: # version + totalLen
					try: 
						(self.hdr.version,self.hdr.totalPackLen,self.hdr.platform,
						self.hdr.frameNumber,self.hdr.timeCpuCycles,self.hdr.numDetectedObj,
						self.hdr.numTLVs,self.hdr.subframeNumber) = struct.unpack('8I', buf[8:40]) 
						self.frameNumber = self.hdr.frameNumber
					except:
						if self.dbg == True:
							print("(Header)Improper TLV structure found: ")
						return (False,'',0)
					
					#self.headerShow()
					
					self.frameNumber = self.hdr.frameNumber
					if self.playback == True:
						return (False,'',0)
					self.JB_syncFlag = 2
					
			#(2) read rest of Data  
			if self.JB_syncFlag == 2:
				r = self.port.read(self.hdr.totalPackLen - self.LEN_MAGIC_HEADER) # read rest of bytes
				return (True,r,self.hdr.numTLVs)
		 
	def rotate_matrix(self,x, y, angle, x_shift=0, y_shift=0, units="DEGREES"):
		# Shift to origin (0,0)
		x = x - x_shift
		y = y - y_shift
		
		# Convert degrees to radians
		if units == "DEGREES":
			angle = np.radians(angle)

		# Rotation matrix multiplication to get rotated x & y
		xr = (x * np.cos(angle)) - (y * np.sin(angle)) + x_shift
		yr = (x * np.sin(angle)) + (y * np.cos(angle)) + y_shift
		return xr, yr

	def tlvRead(self,disp,df=None):
		typeDic = {1000:'V1000',1010:'V1010',1011 :'V1011',1020 : 'V1020',1021 : 'V1021',1040 :'V1040',99: 'idle',999: 'TL'}
		restBuf = b''
		lstate = 'idle'
		
		v1020 = ([]) #v6
		v1010 = ([]) #v7
		v1011 = []   #v8
		v1021 = []
		v1040 = []   #vitalSign Information
		 
		
		idxFrom = 0
		idxTo   = 0
		ttype = 0
		tlvlength = 0
		tlvCnt = 0
		pts = 0
		pc_pts = 0
		while True:
			###############################
			# on {idle} getUartBuf() else parsing from uartBuf[]
			if lstate == 'idle':
				(chk, restBuf,tlvCnt) = self.jb_getUartBuf()
				if chk == True and len(restBuf) != 0:
					lstate = 'TL'
					 
					self.vsdic= {'tid':0,'rangeBin' : 0,
					'breathDev': 0, 'br' : 0  ,'hr':0,
					'hrBuf': [], 'brBuf':[]
					}
					 
					self.vsdic['brBuf'] = []
					self.vsdic['hrBuf'] = []
					
				else:
					return (False,v1020,v1010,v1011,v1021,v1040)
					
			elif lstate == 'TL':
				idxFrom = idxTo; idxTo += 8 #tlv: 8 bytes
				try:
					ttype, tlvlength =  struct.unpack('2I', restBuf[idxFrom:idxTo])
				except:
					print('Exception: TL data unpack')
					
				if ttype in typeDic:
					if self.sm == True:
						print(f"(SM): TL-->{lstate}  len:{tlvlength}")
					lstate = typeDic[ttype]
				else:
					return (False,v1020,v1010,v1011,v1021,v1040)
				
				
			elif lstate == 'V1010':
				idxFrom = idxTo; idxTo += tlvlength
				pts = int( tlvlength / 112)
				if self.sm == True: 
					print(f"(SM): ({lstate})   info: tlvCnt={tlvCnt} tlvlength:{tlvlength} restBuf_len:{len(restBuf)} points:{pts}")
				dataBytes = 112
				
				try:
					for i in range(pts):
						(tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,
						ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi) = struct.unpack('I27f', restBuf[idxFrom + i*dataBytes : idxFrom + (i + 1)*dataBytes])
						
						#posX, posY = self.rotate_matrix(pX, pY, self.azi_offset  , x_shift=0, y_shift=0, units="RADIAN")  
						
						v1010.append((tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi))
				except:
					
					print('Exception: V1010 data unpack')
					return (False,v1020,v1010,v1011,v1021,v1040)

				
				tlvCnt -= 1
				if tlvCnt == 0:
					return (False,v1020,v1010,v1011,v1021,v1040)
				else:
					lstate = 'TL'
					if self.sm == True:
						print(f"(SM): (V1010) --> ({lstate})")
					
			
			elif lstate == 'V1020': # point cloud data
				# Uints
				idxFrom = idxTo; idxTo +=  20   #20 = 5 * 4
				try:
					(elvU,aziU,dopU,ranU,snrU) = struct.unpack('5f', restBuf[idxFrom:idxTo])
					pts = int((tlvlength - 20)/ 8)  # 8: 8types
					pc_pts = pts
					#print(f"(SM): ({lstate})   info: tlvCnt={tlvCnt} tlvlength:{tlvlength} restBuf_len:{len(restBuf)} points:{pts}")
					
					 
					# e,a,d,r,s
					idxFrom = idxTo; idxTo +=  (tlvlength - 20)
					for i in range(pts):
						(e,a,d,r,s) = struct.unpack('2bh2H', restBuf[idxFrom + i*8 : idxFrom + (i + 1)*8]) #8bytes
						elv = e * elvU
						azi = a * aziU
						dop = d * dopU
						ran = r * ranU
						snr = s * ranU
						sz  = ran * np.sin(elv)
						sx  = ran * np.cos(elv) * np.sin(azi)
						sy  = ran * np.cos(elv) * np.cos(azi)
						
						#v1020.append((elv,azi,ran,dop,snr,sx,sy,sz)) 
						v1020.append((sx,sy,sz,dop,snr,azi,elv,ran)) 
				except:
					return (False,v1020,v1010,v1011,v1021,v1040)
					print('Exception: V1020 data unpack')
					
				tlvCnt -= 1
				if tlvCnt == 0:
					return (True,v1020,v1010,v1011,v1021,v1040)
				else:
					lstate = 'TL'
					if self.sm == True:
						print(f"(SM): V1020 --> ({lstate})")
				
			
			elif lstate == 'V1011': #Target Index   ?????????????????????????
				idxFrom = idxTo;  idxTo += tlvlength
				pts = int(tlvlength/1)
				
				try: 
					for i in range(pts):
						ti =  struct.unpack('B', restBuf[idxFrom+i :idxFrom+(i+1)])
						v1011.append(ti[0])
				except:
					print('Exception: V1011 data unpack')
				
				if self.sm == True:
					print(f"(SM): ({lstate})   info: tlvCnt={tlvCnt} tlvlength:{tlvlength} restBuf_len:{len(restBuf)} points:{pts}")
					#print(f"v1011= {v1011}")
				
				tlvCnt -= 1
				if tlvCnt == 0:
					return (True,v1020,v1010,v1011,v1021,v1040)
				else:
					lstate = 'TL'
					if self.sm == True:
						print(f"(SM): V1011 --> ({lstate})")
			
			elif lstate == 'V1021': #Target Presence Indication
				
				idxFrom = idxTo; idxTo += tlvlength
				pts = int(tlvlength/4)
				#print(f"(SM): ({lstate})   info: tlvCnt={tlvCnt} tlvlength:{tlvlength} restBuf_len:{len(restBuf)} points:{pts}")
				try:
					for i in range(pts):
						tpi = struct.unpack('I', restBuf[idxFrom + i*4 : idxFrom + (i + 1)*4])
						v1021.append(tpi[0]) 
				except:
					print('Exception: V1021 data unpack')
					return (False,v1020,v1010,v1011,v1021,v1040)
				 
				tlvCnt -= 1
				if tlvCnt == 0:
					return (True,v1020,v1010,v1011,v1021,v1040)
				else:
					lstate = 'TL'
					if self.sm == True:
						print(f"(SM): V1021 --> ({lstate})")
			
			elif lstate == 'V1040': #vital sign information
				idxFrom = idxTo; idxTo += tlvlength
				try:
					temp = struct.unpack('2H33f',restBuf[idxFrom:idxTo])
					
					self.vs.tid 				= temp[0]
					self.vs.rangeBin			= temp[1]
					self.vs.breathingDeviation  = temp[2]
					self.vs.heart_rate 			= temp[3]
					self.vs.breathing_rate		= temp[4]
					for i in range(15):
						self.vs.vitalSign_heart_buf[i]  = temp[5+i] 
						self.vs.vitalSign_breath_buf[i] = temp[20+i]
						
					self.vsdic= {'tid':self.vs.tid,'rangeBin' : self.vs.rangeBin,
						'breathDev': self.vs.breathingDeviation, 'hr' : self.vs.heart_rate  ,'br':self.vs.breathing_rate,
						'hrBuf': self.vs.vitalSign_heart_buf, 'brBuf':self.vs.vitalSign_breath_buf
						}
					
					#print("tid:{:} rangeBin:{:} bDev:{:.3f} hr:{:.3f} br:{:.3f}".format(self.vs.tid,self.vs.rangeBin,self.vs.breathingDeviation,self.vs.heart_rate,self.vs.breathing_rate))
					#print(f"(SM): ({lstate})   info: tlvCnt={tlvCnt} tlvlength:{tlvlength} restBuf_len:{len(restBuf)} points:{pts}")
					
				except:
					print('Exception: V1041 data unpack')
					return (False,v1020,v1010,v1011,v1021,v1040)
				
				tlvCnt -= 1
				if tlvCnt == 0:
					if self.sm == True:
						print(f"(SM): V1040 --> ESC ")
					return (True,v1020,v1010,v1011,v1021,v1040)
				else:
					lstate = 'TL'
					if self.sm == True:
						print(f"(SM): V1040 --> ({lstate})")
		







# 3D People Counting-ISK/AOP 3D Raw Data for 360°(pc3_360)
# ver:0.0.1
# 2020/06/19
# 
# parsing People Counting 3D fusion for 360° data
# hardware:(Batman-201)ISK IWR6843 ES2.0 / AOP(BM-501/BM601)
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
#===========================================
# output: V6,V7,V8 Raw data
# v1.0.1 : 2022/16/08 release
#          (1)Output list data
#          (2)Output dataframe data
#
# v1.0.2 : 2025/02/11 add zOffset
#          
#  
import serial
import time
import struct
import pandas as pd
import numpy as np

# new
#######################################################
import time
JB_tOld = time.time()
import pandas as pd
pd.options.display.float_format = "{:,.1f}".format
pd.set_option('display.width', 300)
pd.set_option('display.max_columns', 60)
pd.set_option('display.max_rows', 5) # 5 # can see two upper lines and two lower lines and dot line in the middle
def jb(title='.', value=0):
	if 1: # tmp comment out 	
		print('\nJB_({})>'.format(title), end="")
		#print('\nJB_({})>'.format(title)) # ALERT: tmp comment out. due to 'end' 
		if isinstance(value, list):
			print('type= list, len= {}'.format(len(value)))
			print('{}'.format(value))
		elif isinstance(value, tuple):
			print('type = tuple, len= {}'.format(len(value)))
			print('{}'.format(value))
		elif isinstance(value, type(np.array([]))):
			print('type= ndarray, shape= {}'.format(value.shape))
			#print('type= array, len= {}'.format(len(value)))
			print('{}'.format(value))
		elif isinstance(value, type(pd.DataFrame(pd.DataFrame([1.0,2.0])))):
			print('type= dataframe, shape= {}'.format(value.shape))
			print('{}'.format(value))
		elif isinstance(value, type(pd.Series([1.0, 2]))):
			print('type= series, shape= {}'.format(value.shape))
			print('{}'.format(value))
		else:
			print('type={}'.format(type(value)))
			print('JB> value= {}'.format(value))
		print('-----------------------------------------------------------------\n')
####################################################################


class header:
	version = 0
	totalPackLen = 0
	platform = 0
	frameNumber = 0
	subframeNumber = 0
	chirpMargin = 0
	frameMargin = 0 
	trackProcessTime = 0
	uartSendTime = 0
	numTLVs = 0
	checksum = 0

class unitS:
	elevationUnit:float = 0.0
	azimuthUnit : float = 0.0
	dopplerUnit :float = 0.0
	rangeUnit :float = 0.0
	snrUnit :float = 0.0

class Pc3_360:
	gdata = b''
	playback = False
	magicWord =  [b'\x02',b'\x01',b'\x04',b'\x03',b'\x06',b'\x05',b'\x08',b'\x07',b'\0x99']
	port = ""
	hdr = header
	u = unitS
	frameNumber = 0
	# provide csv file dataframe
	# real-time 
    #v6_col_names_rt = ['time','fN','type','elv','azimuth','range' ,'doppler','snr','sx', 'sy', 'sz'] # original
	v6_col_names_rt = ['fN','type','elv','azimuth','doppler','range' ,'snr',  'sx', 'sy', 'sz' ]  #v0.1.4
	v7_col_names_rt = ['fN','type','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','ec0','ec1','ec2','ec3','ec4','ec5','ec6','ec7','ec8','ec9','ec10','ec11','ec12','ec13','ec14','ec15','g','confi','tid']
	                     
	v8_col_names_rt = ['fN','type','targetID']
	zOffset = 0.0 
	
	# read from file for trace point clouds
	fileName = ''
	v6_col_names = ['time','fN','degree','type','elv','azimuth','doppler','range','snr','sx', 'sy', 'sz']
	v7_col_names = ['time','fN','degree','type','tid','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','ec0','ec1','ec2','ec3','ec4','ec5','ec6','ec7','ec8','ec9','ec10','ec11','ec12','ec13','ec14','ec15','g','confi'] #v0.1.2
	v8_col_names = ['time','fN','degree','type','targetID']
	
	v6simo = []
	v7simo = []
	v8simo = []
	
	sim_startFN = 0
	sim_stopFN  = 0 
	JB_t0 = 0
	version_keep = '3050004'
	
	# add for interal use
	tlvLength = 0
	numOfPoints = 0
	# for debug use 
	dbg = False #Packet unpacket Check: True show message 
	sm = False #Observed StateMachine: True Show message
	plen = 16 
	
	def __init__(self,port, azi_degree = None, bufSize = None, zOffset = None):
		
		port.reset_input_buffer()
		port.ReadBufferSize = bufSize
		self.degree = azi_degree
		self.port = port
		self.zOffset = 0.0 if zOffset == None else zOffset
		self.azi_offset = 0.0 if azi_degree == None else azi_degree * np.pi/180.0
		print("(jb)People Counting 3D initial")
		print("(jb)vsersion:v0.1.0")
		print("(jb)For Hardware:Batman-201(ISK)/AOP(501/601)")
		print("(jb)Firmware: PC3")
		print("(jb)UART Baud Rate:921600")
		print("Output: V6,V7,V8 data:(RAW)\n##########################################\n\n\n")
		
	def useDebug(self,ft):
		self.dbg = ft
		
	def stateMachine(self,ft):
		self.sm = ft
		
	def getHeader(self):
		return self.hdr
		
	def headerShow(self):
		print("***header***********") 
		print("Version:     \t%x "%(self.hdr.version))
		print("Platform:    \t%X "%(self.hdr.platform))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
		print("PID(frame#): \t%d "%(self.hdr.frameNumber))
		print("subframe#  : \t%d "%(self.hdr.subframeNumber))
		print("Inter-frame Processing Time:\t{:d} us".format(self.hdr.trackProcessTime))
		print("UART Send Time:\t{:d} us".format(self.hdr.uartSendTime))
		print("Inter-chirp Processing Margin:\t{:d} us".format(self.hdr.chirpMargin))
		print("Inter-frame Processing Margin:\t{:d} us".format(self.hdr.frameMargin))
		print("numTLVs:     \t%d "%(self.hdr.numTLVs))
		print("Check Sum   :\t{:x}".format(self.hdr.checksum))
		print("***End Of Header***") 
			
	
		
	def list2df(self,dck,l6,l7,l8):
		ll6 = pd.DataFrame(l6,columns=self.v6_col_names_rt)
		ll7 = pd.DataFrame(l7,columns=self.v7_col_names_rt)
		ll8 = pd.DataFrame(l8,columns=self.v8_col_names_rt)
		return (dck,ll6,ll7,ll8)

#
# TLV: Type-Length-Value
# read TLV data
# input:
#     disp: True:print message
#			False: hide printing message
#     df: 
# output:(return parameter)
# (pass_fail, v6, v7, v8)
#  pass_fail: True: Data available    False: Data not available
#  v6: point cloud infomation
#  v7: Target Object information
#  v8: Target Index information
#  
# 
#

	#################################################################
	JB_syncFlag = 0
	def jb_getUartBuf(self):
		idx = 0
		buf = b""
		self.gdata = b''
		while True:
			if self.JB_syncFlag == 0:
				ch = self.port.read() # here ch type is byte
				if ch == self.magicWord[idx]: # 0..7
					buf += ch
					idx += 1 # 1..8
					if idx == 8:
						idx = 8
						self.JB_syncFlag = 1
						self.JB_t0 = time.time()
						
			else:
				self.JB_syncFlag = 0 # magic word ch mismatch loop back again
				idx = 0 # init
				buf = b""
				self.gdata = b''
			

			if self.JB_syncFlag == 1:
				# read next 16 bytes for v, l, p, f
				for i in range(16):
					ch = self.port.read()
					
					buf += ch
					self.gdata = buf
					
				if len(buf) == (8 + 16): # version + totalLen
					magic=struct.unpack('8B', buf[0:8])
					version = struct.unpack('I', buf[8:12])[0]   # get tuple [0]
					verString = "{:X}".format(version)
					totalLen = struct.unpack('I', buf[12:16])[0] # get tuple [0]
					fn = struct.unpack('I', buf[20:24])[0]
					
					self.frameNumber = fn
					if self.playback == True:
						return False
					
					if verString != self.version_keep or totalLen > 100000:
						print("ver: {:}   version_keep:{:}".format(verString,self.version_keep ))
						return False
					
					
				else:
					self.JB_syncFlag = 0 # len mismatch loop back again
					idx = 0
					buf = b''
					self.gdata = b''

			if self.JB_syncFlag == 1:
				r = self.port.read(totalLen - 24) # read rest of bytes
				#xx = self.port.inWaiting()
				#print("is Wating= {:}  :len:{:}".format(0,len(r)))
				#buf += r
				self.gdata += r
				dt = time.time()- self.JB_t0
				#self.port.flushInput()
				self.port.reset_input_buffer()
				#print('JB_getUartBuf> fn={}, magic={}, version={:X}, totalLen={}, bufLen={}, dt={:.0f} ms'.format(fn, magic, version, totalLen, len(buf), dt * 1000))
				if totalLen == len(self.gdata):
					return True
				else:
					self.JB_syncFlag = 0 # len mismatch loop back again
					idx = 0
					buf = b""
					self.gdata = b''
					

#################################################################
#
# TLV: Type-Length-Value
# read TLV data
# input:
#     disp: True:print message
#			False: hide printing message
#     df: 
# output:(return parameter)
# (pass_fail, v6, v7, v8)
#  pass_fail: True: Data available    False: Data not available
#  v6: point cloud infomation
#  v7: Target Object information
#  v8: Target Index information
#  
#================================================================
	JB_i = 0
	def tlvRead(self,disp,df = None):
		global JB_tOld, JB_i
		 
		#print("---tlvRead---")
		#ds = dos
		typeList = [6,7,8]
		idx = 0
		lstate = 'idle'
		sbuf = b''
		lenCount = 0
		unitByteCount = 0
		dataBytes = 0
		numOfPoints = 0
		tlvCount = 0
		pbyte = 16
		v6 = ([])
		v7 = ([])
		v8 = ([])
		v6df = ([])
		v7df = ([])
		v8df = ([])
		ch = b'\x00' # init
		idxTo = 0
		uartBuf = b''
		while True:
			###############################
			# on {idle} getUartBuf() else parsing from uartBuf[]
			if lstate == 'idle':
				chk = self.jb_getUartBuf()
				if chk == True:
					uartBuf = self.gdata
				else:
					#print("===========================tlvRead=========================buf empty" )
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
		
				self.JB_i = 8 # init, read start from 8, after magic word		 
				lstate = 'header'
				sbuf = b"" # init
				idxFrom = 0
				idxTo = 8
			else:
				dumy = 0
				
				# read ch from uartBuf[]
				#ch1 = uartBuf[self.JB_i] # read start after magic word (8 bytes)
				#self.JB_i += 1 		
				# MUST int -> bytes			
				#ch = bytes(ch1) # ALERT: failed
				#ch = ch1.to_bytes(1, 'little')
			################################

			if lstate == 'header': # 40B := index 8:48
				idxFrom = idxTo
				idxTo += 40
				#print("JB> header : idxTo:{:}".format(idxTo))
				try: 
					(self.hdr.version,self.hdr.totalPackLen,self.hdr.platform,
					self.hdr.frameNumber,self.hdr.subframeNumber,
					self.hdr.chirpMargin,self.hdr.frameMargin,self.hdr.trackProcessTime,self.hdr.uartSendTime,self.hdr.numTLVs,self.hdr.checksum
					) = struct.unpack('9I2H', uartBuf[idxFrom : idxTo]) 
					self.frameNumber = self.hdr.frameNumber
					
					#self.headerShow()
					#print('JB> fn={}, len={}, platform={}, tlv={}'.format(self.frameNumber, self.hdr.totalPackLen, self.hdr.platform, self.hdr.numTLVs))
					if self.playback == True:
						lstate = 'idle'
					else:
						lstate = 'TL'
				except:
					if self.dbg == True:
						print("(Header)Improper TLV structure found: ")
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
				
				tlvCount = self.hdr.numTLVs
					
			elif lstate == 'TL': #TLV Header type/length
				idxFrom = idxTo
				idxTo += 8
				
				#print("JB> (TL) indexFrom = {:}   idxTo = {:}".format(idxFrom,idxTo))
				#print("JB> (TL) {:}".format(uartBuf[idxFrom : idxTo]))
				ttype, self.tlvLength = struct.unpack('2I', uartBuf[idxFrom : idxTo])
				
				if ttype == 6 or ttype == 7:
					#print('JB> (TL) type=V{}, len={}'.format(ttype, self.tlvLength))
					if ttype  == 6: #'V6'
						lstate = 'V6-unit' #'V6'
						if self.sm == True:
							print("JB>..(TL:len={:d})=>({:})".format(self.tlvLength,lstate))
							
					if ttype == 7:
						return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
						lstate = 'idle'
						#lstate = 'V7'
						if self.sm == True:
							print("JB>..(TL:len={:d})=>({:})".format(self.tlvLength,lstate))
				else:
					print("(TL) -v6/v7 failed  type: {:}".format(ttype))
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
			
			elif lstate == 'V6-unit':
				idxFrom = idxTo
				idxTo +=  (5 * 4)    #(self.tlvLength - 8) # uints := 5 x u32
				#print(len(uartBuf[idxFrom  : idxTo]))
				try: 
					self.u.elevationUnit,self.u.azimuthUnit,self.u.dopplerUnit,self.u.rangeUnit,self.u.snrUnit = struct.unpack('5f', uartBuf[idxFrom : idxTo])
					#print("JB> (V6-unit)  ==> elv:{:.4f} azimuth:{:.4f} doppler:{:.4f} range:{:.4f} snr:{:.4f}".format(self.u.elevationUnit,self.u.azimuthUnit,self.u.dopplerUnit,self.u.rangeUnit,self.u.snrUnit))
				except:
					if self.dbg == True:
						print("(V6-unit)Improper Type/Value structure found: ")
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
					
				lstate = 'V6'
				if self.sm == True:
					print("JB>..(V6-Unit)=>(V6)")
					
			elif lstate == 'V6': # count = Total Lentgh - 8   # eadrs: 8 bytes
				dataBytes = 8
				datalen = self.tlvLength - 8 - 20
				points = int(datalen/dataBytes)
				#print("JB> (V6) points: {:}    datalen:{:}  idxTo:{:}".format(points,datalen,idxTo))
				idxFrom = idxTo
				idxTo += datalen 
				#print("JB> (V6) idxFrom:{:} idxTo= {:}  diff:{:} + 28 = {:}    self.tlvLength= {:}".format(idxFrom,idxTo,idxTo-idxFrom,idxTo-idxFrom + 28,self.tlvLength))
				
				#sbuf = uartBuf[idxFrom : idxTo]
				for i in range(points):
					try: 
						(e,a,d,r,s) = struct.unpack('2bh2H', uartBuf[idxFrom + i*dataBytes : idxFrom + (i + 1)*dataBytes]) #8bytes
						elv = e * self.u.elevationUnit
						azi = a * self.u.azimuthUnit  + self.azi_offset
						dop = d * self.u.dopplerUnit
						ran = r * self.u.rangeUnit
						snr = s * self.u.snrUnit
						sz  = ran * np.sin(elv)  + self.zOffset
						sx  = ran * np.cos(elv) * np.sin(azi) 
						sy  = ran * np.cos(elv) * np.cos(azi) 
						
						#print("({:}:{:4d})(idx:({:4d}) elv:{:.4f} azimuth:{:.4f} doppler:{:.4f} range:{:.4f} snr:{:.4f}".format(points,lenCount,idx,elv,azi,dop,ran,snr))
						if (df == 'DataFrame'):
							v6df.append((self.hdr.frameNumber,'v6',elv,azi,dop,ran,snr,sx,sy,sz))
							#jb('tlvRead()_v6df', v6df)
						else:
							v6.append((elv,azi,dop,ran,snr,sx,sy,sz))
					except:
						if self.dbg == True:
							print("(7)Improper Type 6 Value structure found: ")
						return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
				
				tlvCount -= 1
				if tlvCount == 0:
					lstate = 'idle'
					if self.sm == True:
						print("JB>..(V6)->(idle)   tlvCount:{:}".format(tlvCount))
					return self.list2df(True,v6df,v7df,v8df) if (df == 'DataFrame') else (True,v6,v7,v8)
				else:
					lstate = 'TL'
					if self.sm == True:
						print("JB>..(V6)->(TL)  tlvCount:{:}".format(tlvCount))
						
			elif lstate == 'V7': # count = Total Lentgh - 8
				dataBytes = 112 # 4 + 9 * 4 + 18 *4 = 112
				datalen = self.tlvLength - 8
				points = int(datalen / dataBytes)
				idxFrom = idxTo
				idxTo += datalen
				#print("JB> (V7) points: {:}    datalen:{:}  idxTo:{:}".format(points,datalen,idxTo))
				#print("JB> (V7) idxFrom: {:}  idxTo:{:}".format(idxFrom,idxTo))
				
				for i in range(points):
					try:
						(tid,pX,pY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,
						ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi) = struct.unpack('I9f18f', uartBuf[idxFrom + i*dataBytes : idxFrom + (i + 1)*dataBytes])
						posX, posY = rotate_matrix(pX, pY, self.azi_offset  , x_shift=0, y_shift=0, units="RADIAN")  
						
						if (df == 'DataFrame'):
							v7df.append((self.hdr.frameNumber,'v7',posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,
							ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi,tid)) 
						else: 
							v7.append((tid,posX,posY,posZ,velX,velY,velZ,accX,accY,accZ,ec0,ec1,ec2,ec3,ec4,ec5,ec6,ec7,ec8,ec9,ec10,ec11,ec12,ec13,ec14,ec15,g,confi))
							
					except:
						if self.dbg == True:
							print("(7)Improper Type 7 Value structure found: ")
						return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)
				
				tlvCount -= 1
				if tlvCount == 0:
					lstate = 'idle'
					if self.sm == True:
						print("JB> (V7)->(idle)   tlvCount:{:}".format(tlvCount))
					return self.list2df(True,v6df,v7df,v8df) if (df == 'DataFrame') else (True,v6,v7,v8)
				else:
					lstate = 'TL'
					if self.sm == True:
						print("JB> (V7)->(TL)  tlvCount:{:}".format(tlvCount))
						
			elif lstate == 'V8':
				idx += 1
				v8.append(ord(ch))
				
				if idx == lenCount:
					if disp == True:
						print("=====V8 End====")
					sbuf = b""
					idx = 0
					lstate = 'idle'
					if self.sm == True:
						print("(V8:{:d})=>(idle)".format(tlvCount))
						
					if (df == 'DataFrame'):
						v8o = [self.hdr.frameNumber,'v8']
						v8df = v8o.extend(v8)
					return self.list2df(True,v6df,v7df,v8df) if (df == 'DataFrame') else (True,v6,v7,v8)
				
				if idx > lenCount:
					sbuf = b""
					idx = 0
					lstate = 'idle'
					return self.list2df(False,v6df,v7df,v8df) if (df == 'DataFrame') else (False,v6,v7,v8)

	def rotate_matrix(x, y, angle, x_shift=0, y_shift=0, units="DEGREES"):
		# Shift to origin (0,0)
		x = x - x_shift
		y = y - y_shift
		
		# Convert degrees to radians
		if units == "DEGREES":
			angle = np.radians(angle)

		# Rotation matrix multiplication to get rotated x & y
		xr = (x * np.cos(angle)) - (y * np.sin(angle)) + x_shift
		yr = (x * np.sin(angle)) + (y * np.cos(angle)) + y_shift
		return xr, yr


					

	def v67Simlog(frameNum):
		global sim_startFN,sim_stopFN
		s_fn = frameNum + sim_startFN
		#print("frame number:{:}".format(s_fn))
		v6d = v6sim[v6sim['fN'] == s_fn]
		#v6d =  v6dd[v6dd['doppler'] < 0.0]
		#print(v6d)
		v7d = v7sim[v7sim['fN'] == s_fn]
		chk = 0
		if v6d.count != 0:
			chk = 1
		return (chk,v6d,v7d)
		
	def getRecordData(self,frameNum):
		s_fn = frameNum + self.sim_startFN
		#print("frame number:{:}".format(s_fn))
		v6d = self.v6simo[self.v6simo['fN'] == s_fn]
		#v6d =  v6dd[v6dd['doppler'] < 0.0]
		#print(v6d)
		v7d = self.v7simo[self.v7simo['fN'] == s_fn]
		v8d = self.v8simo[self.v8simo['fN'] == s_fn]
		chk = 0
		if v6d.count != 0:
			chk = 1
		return (chk,v6d,v7d,v8d)
		
	
		
	def readFile(self,fileName):
		#fileName = "pc32021-03-19-10-02-17.csv"  
		#df = pd.read_csv(fileName, error_bad_lines=False, warn_bad_lines=False) 
		self.fileName = fileName 
		#          ['time','fN','type','elv','azimuth','range' ,'doppler','snr','sx', 'sy', 'sz']
		df = pd.read_csv(self.fileName, names = self.v6_col_names, skiprows = [0,11,12]) 
		df.dropna()
		#print("------------------- df --------------------shape:{:}".format(df.shape))
		print(df.info())
		print(df.info(memory_usage="deep")) 
		
		v6simOri = df[(df.type == 'v6')]
		#print("-------------------v6sim------------:{:}".format(v6simOri.shape))
		
		self.v6simo = v6simOri.loc[:,['fN','type','elv','azimuth','doppler','range' ,'snr','sx', 'sy', 'sz']] # in this case
		self.v6simo['elv'] = self.v6simo['elv'].astype(float, errors = 'raise') 
		
		df7 = pd.read_csv(self.fileName, names = self.v7_col_names, skiprows = [0])  
		
		#------------- v7 sim ---------------
		v7simc = df7[df7['type'] == 'v7']
		self.v7simo  = v7simc.loc[:,['fN','type','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','tid']]
		self.v7simo['posX'] = self.v7simo['posX'].astype(float, errors = 'raise') 
		self.sim_startFN = df['fN'].values[0]
		self.sim_stopFN  = df['fN'].values[-1]
		
		#------------- v8 sim ---------------
		v8simc = df[df['type'] == 'v8']
		self.v8simo  = v8simc.loc[:,['fN','type','elv']]
		self.v8simo.columns = ['fN','type','targetID']
		
		#print(self.v8simo)
		return (self.v6simo,self.v7simo,self.v8simo)
		







# Obstacle Detection Sensing
# 2023/06/18:  
#  
# hardware: BM-501 
#     
# company: Joybien Technologies: www.joybien.com
# author: Zach Chen
# v0.0.1
#===========================================
# output: V17,V89,V10
# 

import serial
import time
import struct
import pandas as pd
import numpy as np

class header:
	version = 0
	totalPackLen = 0
	platform = 0
	frameNumber = 0
	timeInCPUcycle = 0
	numDetectedObj = 0
	numTLVs = 0
	subframeNumber = 0
	numStaticDetectedObj = 0
	
class ObstacleDS:
	magicWordR =  [b'\x02',b'\x01',b'\x04',b'\x03',b'\x06',b'\x05',b'\x08',b'\x07',b'\0x99']  #ori
	magicWord  =  [b'\x01',b'\x02',b'\x03',b'\x04',b'\x05',b'\x06',b'\x07',b'\x08',b'\0x99']  #use
	
	
	typeList = [1,7,8,9,10,11] 
	X_CALI_DEGREE = 4.85
	INSTALL_HEIGHT = 0
	V_COMBINE = True
	port = ""
	hdr = header
	r = False
	
	# provide csv file dataframe
	# real-time 
	v1_col_names_rt   = ['fN','type','range','elv','angle','doppler','sx','sy','sz']
	v7_col_names_rt   = ['fN','type','snr','noise']
	v8_col_names_rt   = ['fN','type','X','Y','Z','doppler']
	v9_col_names_rt   = ['fN','type','snr','noise']
	v10_col_names_rt  = ['fN','type','tid','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ']
	v11_col_names_rt  = ['fN','type','Association']
	v17_col_names_rt =  ['fN','type','range','elv','angle','doppler','sx','sy','sz','snr','noise','Association']
	v89_col_names_rt =  ['fN','type','X','Y','Z','doppler','snr','noise']
	
	vxx_col_names = ['fN','type','i0','i1','i2','i3','i4', 'i5','i6' ,'i7','i8','i9']  
	# read from file for trace point clouds
	fileName = ''
	
	sim_startFN = 0
	sim_stopFN  = 0 
	
	# add for interal use
	tlvLength = 0
	numOfPoints = 0
	# for debug use 
	dbg = False #Packet unpacket Check: True show message 
	sm  = False #Observed StateMachine: True Show message
	plen = 16 
	
	def __init__(self,port):
		
		self.port = port
		print("\n\n\n(jb)mmWave ODS initial")
		print("(jb)version:v0.1.0")
		print("(jb)For Hardware:BM-501")
		print("(jb)Hardware: BM501")
		print("(jb)Firmware:")
		print("(jb)UART Baud Rate:921600")
		print("(jb)Output: V17,V89 and V10 data:(RAW)\n")
		print('v17: Detected Object(Dynamic Points\t\nv89: Static Detected Object(Static Points)')
		print('v10: Tracked Object\n\n')
		
		
	def useDebug(self,ft):
		self.dbg = ft
		
	def stateMachine(self,ft):
		self.sm = ft
		
	def getHeader(self):
		return self.hdr
		
	
	def headerShow(self):
		print("\n\n\n******* Header ********") 
		print("Version:     \t%x "%(self.hdr.version))
		print("TotalPackLen:\t%d "%(self.hdr.totalPackLen))
		print("Platform:    \t%X "%(self.hdr.platform))
		print("PID(frame#): \t%d "%(self.hdr.frameNumber))
		print("time in CPU cycles:\t{:d} us".format(self.hdr.timeInCPUcycle))
		print("number of Detected object:\t{:d}".format(self.hdr.numDetectedObj))
		print("numTLVs:     \t%d "%(self.hdr.numTLVs))
		print("subframe#  : \t%d "%(self.hdr.subframeNumber))
		print("number of staic Detected object:\t%d "%(self.hdr.numStaticDetectedObj))
		print("***End Of Header***\n") 
			

	#for class internal use
	def tlvTypeInfo(self,dtype,count,dShow):
		
		sbyte = 8  #tlvHeader Struct = 8 bytes
		unitByte = 20 
		dataByte = 2
		pString = ""
		nString = "numOfPoints :"
		stateString = "Vx-unit"
		retCnt = count
		nPoint = 0
		
		if dtype == 1:  #chk ok ods
			unitByte = 0  
			dataByte= 16 #point bytes= 16bytes (range(4),angle(4),elv(4),doppler(4))
			pString = "Detected Object (Dynamic Points)"
			stateString = "V1"
			nPoint = count/dataByte
			retCnt = count
		
		elif dtype == 7:  #chk ok ods
			unitByte = 0  #pointUnit= 0
			sbyte = 8      #tlvHeader Struct = 8 bytes (type(4)/lenth(4))
			dataByte= 4    #pointStruct 4bytes:(snr(2),noise(2))
			pString = "Detected Object (Dynamic Points) Side Info"
			nPoint = count/dataByte
			retCnt = count
			stateString = "V7"
		
		elif dtype == 8: #chk ok ods
			unitByte = 0  
			dataByte= 16 #point bytes= 16bytes (X(4),Y(4),Z(4),doppler(4))
			pString = "Static Detected Object (Static Points)"
			stateString = "V8"
			nPoint = count/dataByte
			retCnt = count
			
		elif dtype == 9: #chk ok ods
			unitByte = 0  #pointUnit= 0
			sbyte = 8      #tlvHeader Struct = 8 bytes (type(4)/lenth(4))
			dataByte= 4    #pointStruct 4bytes:(snr(2),noise(2))
			pString = "Static Detected Object (Static Points) Side Info"
			nPoint = count/dataByte
			retCnt = count
			stateString = "V9"
			
		elif dtype == 10: #chk ok ods
			unitByte = 0  #pointUnit= 0
			sbyte = 8      #tlvHeader Struct = 8 bytes (type(4)/lenth(4))
			dataByte= 40    #pointStruct 40bytes:(snr(2),noise(2))
			pString = "Tracked Object List"
			nPoint = count/dataByte
			retCnt = count
			stateString = "V10"
			
		elif dtype == 11: #point to Track Assocition
			unitByte = 0  #pointUnit= 0
			sbyte = 8      #tlvHeader Struct = 8 bytes (type(4)/lenth(4))
			dataByte= 1    #pointStruct 40bytes:(snr(2),noise(2))
			pString = "Point to Track Assocition"
			nPoint = count/dataByte
			retCnt = count
			stateString = "V11"

		else:
			unitByte = 0
			sbyte = 1
			pString = "*** Type Error ***"
			stateString = 'idle' 

		#dShow = True
		if dShow == True:
			print("-----[{:}] ----".format(pString))
			print("tlv Type({:2d}Bytes):  \tv{:d}".format(sbyte,dtype))
			print("tlv length:      \t{:d}".format(count)) 
			print("{:}      \t{:d}".format(nString,int(nPoint)))
			print("value length:    \t{:d}".format(retCnt))
			print("StateString= {}".format(stateString))  
			print("---------------------")
		
		return unitByte,stateString, sbyte, dataByte,retCnt, int(nPoint)
		
	def list2df(self,dck,l1,l7,l8,l9,l10,l11):
		ll1  = pd.DataFrame(l1,columns=self.v1_col_names_rt)
		ll7  = pd.DataFrame(l7,columns=self.v7_col_names_rt)
		ll8  = pd.DataFrame(l8,columns=self.v8_col_names_rt) 
		ll9  = pd.DataFrame(l9,columns=self.v9_col_names_rt) 
		ll10 = pd.DataFrame(l10,columns=self.v10_col_names_rt) 
		ll11 = pd.DataFrame(l11,columns=self.v11_col_names_rt) 
		 
		return (dck,ll1,ll7,ll8,ll9,ll10,ll11)  
	'''
	v10_col_names_rt  = ['fN','type','tid','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ']
	v11_col_names_rt  = ['fN','type','Association']
	v17_col_names_rt =  ['fN','type','range','elv','angle','doppler','sx','sy','sz','snr','noise','Association']
	v89_col_names_rt =  ['fN','type','X','Y','Z','doppler','snr','noise']
	'''
	def list2df_(self,dck,l1,l7,l8,l9,l10,l11):
		
		ll17s = [v[:1] +('v17',) + v[2:] + a[2:] + b[2:] for v, a, b in zip(l1, l7, l11)]
		#ll89s = [v + a[2:] for v, a  in zip(l8, l9)]
		ll89s = [v[:1] +('v89',) + v[2:] + a[2:] for v, a  in zip(l8, l9)]
		 
		v17o   = pd.DataFrame(ll17s,columns=self.v17_col_names_rt) 
		v89o  = pd.DataFrame(ll89s,columns=self.v89_col_names_rt) 
		v10o  = pd.DataFrame(l10,columns=self.v10_col_names_rt) 
		 
		return (dck,v17o,v89o,v10o)  

	def x_calibrate(self,x,y):
		return  x + np.array(y) * np.tan(np.deg2rad(self.X_CALI_DEGREE))
		
		
	def v17Combine(self,v1 = None, v7 = None, v11 = None):
		v17 = [(*t1,  *t7) for t1,  t7 in zip(v1, v7)] 
		v17 = [t + (item,) for t, item in zip(v17, v11)]
		return v17
	
	def v89Combine(self,v8 = None, v9 = None):
		return [(*t8, *t9) for t8, t9 in zip(v8, v9)]
		
	def combine(self, dck,v1,v7,v8,v9,v10,v11):
		if self.V_COMBINE:
			v89 = self.v89Combine(v8 = v8, v9 = v9)
			v17 = self.v17Combine(v1 = v1, v7 =v7, v11 =v11)
			return (dck,v17,v89,v10)
		else:
			return (dck,v1,v7,v8,v9,v10,v11)
			
	stateR = 0
	stateF = 0
	def state_machineR(self,byte, magic_word):
		print(f"R:input= {byte} magic_word[{self.stateR}] = {magic_word[self.stateR]}   " ) 
		if byte == magic_word[self.stateR]:
			self.stateR += 1
			if self.stateR == len(magic_word):
				self.stateR = 0  # Reset the state machine
				return True
		else:
			self.stateR = 0  # Reset the state if the byte doesn't match the expected pattern
		return False
	
	def state_machineF(self,byte, magic_word): 
		print(f"B:input= {byte} magic_word[{self.stateF}] = {magic_word[self.stateF]}   " ) 
		if byte == magic_word[self.stateF]:
			self.stateF += 1
			if self.stateF == len(magic_word):
				self.stateF = 0  # Reset the state machine
				return True
		else:
			self.stateF = 0  # Reset the state if the byte doesn't match the expected pattern
		return False
#
# TLV: Type-Length-Value
# read TLV data
# Usuage:
# (dck,v1,v7,v8,v9,v10,v11)  = radar.tlvRead(True,df = 'DataFrame')
# input:
#     disp: True:print message
#	  False: hide printing message
#  
# output:(return parameter)
# (pass_fail,v1,v7,v8,v9,v10,v11)
#  pass_fail: True: Data available    False: Data not available
#  v1: Detected Object(Dynamic Points
#  v7: Detected Object(Dynamic Points) Side Info 
#  v8: Static Detected Object(Static Points) 
#  v9: Static Detected Object(Static Points) Side Info
#  v10: Tracked Object List
#  v11: Point to Track Association 
#
 
	def tlvRead(self,disp,df = None ):
		self.magicWord = self.magicWordR if self.r else self.magicWord
		idx = 0
		lstate = 'idle'
		sbuf = b""
		lenCount = 0
		unitByteCount = 0
		escapeCount = 0
		dataBytes = 0
		numOfPoints = 0
		numOfPointsCnt = 0
		tlvCount = 0
		pbyte = 16
		
		v1 = ([])
		v7 = ([])
		v8 = ([])
		v9 = ([])
		v10 = ([])
		v11 = ([])
		
		v17 = ([])
		v89 = ([])
		
		v1df = ([])
		v7df = ([])
		v8df = ([])
		v9df = ([])
		v10df = ([])
		v11df = ([])
		
		v17simo = []
		
		nextState = False
		while True:
			try:
				ch = self.port.read()
			except:
				return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
				
			if lstate == 'idle':
				#print(f"ch = {ch}  magicWord[{idx}]= {self.magicWord[idx]}     state={lstate}")
				
				if ch == self.magicWord[idx]:
					idx += 1
					if idx == 8:
						#print("=================== magic word ==================================")
						idx = 0
						lstate = 'header'
						rangeProfile = b""
						sbuf = b""
						self.escapeCount = 0
				else:
					#print("not: magicWord state:")
					idx = 0
					rangeProfile = b""
					return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
				
			elif lstate == 'header':
				 
				sbuf += ch
				idx += 1
				escapeCount += 1
				if idx == 36: 
					#print("------header-----")
					#print(":".join("{:02x}".format(c) for c in sbuf)) 	 
					#print("len:{:d}".format(len(sbuf))) 
					# [header - Magicword]
					
					try: 
						(self.hdr.version,self.hdr.totalPackLen,
						self.hdr.platform,self.hdr.frameNumber,
						self.hdr.timeInCPUcycle,self.hdr.numDetectedObj,
						self.hdr.numTLVs,self.hdr.subframeNumber,self.hdr.numStaticDetectedObj) = struct.unpack('9I', sbuf)
						self.frameNumber = self.hdr.frameNumber
					except:
						if self.dbg == True:
							print("(Header)Improper TLV structure found: ")
						return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
					
					if disp == True:  
						self.headerShow()
					
					tlvCount = self.hdr.numTLVs
					#print("tlvCount:{:}".format(tlvCount))
					if self.hdr.numTLVs == 0:
						return self.list2df(True,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(True,v1,v7,v8,v9,v10,v11)
						
					if self.sm == True:
						print("(Header)")
						
					sbuf = b""
					idx = 0
					lstate = 'TL'
					  
				elif idx > 64: #44
					idx = 0
					sbuf = b''
					lstate = 'idle'
					return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
					
			elif lstate == 'TL': #TLV Header type/length
				sbuf += ch
				idx += 1
				self.escapeCount += 1
				
				if idx == 8:
					#print(":".join("{:02x}".format(c) for c in sbuf))
					try:
						ttype,self.tlvLength = struct.unpack('2I', sbuf)
						if self.dbg == True:
							print("-------------------------")
							print("(TL)--tlvNum:{:d}: tlvCount({:d})-------ttype:tlvLength:{:d}:{:d}".format(self.hdr.numTLVs,tlvCount,ttype,self.tlvLength))
						if ttype not in self.typeList or self.tlvLength > 20000:
							if self.dbg == True:
								print("(TL)Improper TL Length(hex):(T){:d} (L){:x} numTLVs:{:d}".format(ttype,self.tlvLength,self.hdr.numTLVs))
							sbuf = b""
							idx = 0
							lstate = 'idle'
							self.port.flushInput()
							return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
							
					except:
						if self.dbg == True:
							print("TL unpack Improper Data Found:")
						self.port.flushInput()
						return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
						
					
						
					unitByteCount,lstate ,plen ,dataBytes,lenCount, numOfPoints = self.tlvTypeInfo(ttype,self.tlvLength,disp)
					
					if self.dbg == True:
						print("(TL) unitByteCount={}  lstate ={} , bytes(type+len)={} dataBytes= {} lenCount ={} numOfPoint={}".format(unitByteCount,lstate ,plen ,dataBytes,lenCount, numOfPoints))
					#if ttype == 10:
					#	print("--pointCloud:((tlvLength({:d})-pointUnit(20)-tlvStruct(8))/8={:d}".format(self.tlvLength,numOfPoints))
					if self.sm == True:
						print("(TL:tlvCount:{:d})=>({:})".format(tlvCount,lstate))
					
					tlvCount -= 1
					#print("(TL->type:{:})--tlvNum:{:d}: tlvCount({:d})".format(ttype,self.hdr.numTLVs,tlvCount))
					idx = 0  
					sbuf = b""
					numOfPointsCnt = 0
					
					if ttype == 1 and self.tlvLength == 0 and tlvCount != 0:
						if self.dbg == True:
							print("========type=1 lenth = 0 ===========================================")
						idx = 0
						lstate = 'idle'
						
					if self.escapeCount > self.hdr.totalPackLen:
						if self.dbg == True:
							print("(ESC)========type={:} ========self.escapeCount > self.hdr.totoalPackLen==============================".format(lstate))
						idx = 0
						sbuf = b''
						lstate = 'idle'
					
					
			elif lstate == 'V1' or lstate == 'V8' or lstate == 'V10': #v1 chk ok 
				sbuf += ch
				idx += 1
				self.escapeCount += 1
				if (idx%dataBytes == 0):
					try:
						#print(":".join("{:02x}".format(c) for c in sbuf))
						#v1:raed
						#v8:XYZD
						if lstate == 'V1' or lstate == 'V8':
							(r,a,e,d) = struct.unpack('4f', sbuf)
							if lstate == 'V8':
								if self.dbg == True:
									print("(V8)=> (npts={:2d}: lenCnt={:4d})(idx:({:4d}) X:{:.4f} Y:{:.4f} Z:{:.4f} d:{:.4f}".format(numOfPoints,lenCount,idx,r,a,e,d))
								if (df == 'DataFrame'):
									v8df.append((self.hdr.frameNumber,'v8',r,a,e + self.INSTALL_HEIGHT,d)) 
								else:
									v8.append((r,a,e + self.INSTALL_HEIGHT,d))
							
							if lstate == 'V1':
								
								sz = r * np.sin(e)
								sx  = r * np.cos(e) * np.sin(a)
								sy  = r * np.cos(e) * np.cos(a)
								 
								
								if self.dbg == True:
									print("(V1)=> (npts={:2d}: lenCnt={:4d})(idx:({:4d}) r:{:.4f} e:{:.4f} a:{:.4f} d:{:.4f}".format(numOfPoints,lenCount,idx,r,e,a,d,sx,sy,sz))
								if (df == 'DataFrame'):
									v1df.append((self.hdr.frameNumber,'v1',r,e,a,d,sx,sy,s + self.INSTALL_HEIGHTz)) 
								else:
									v1.append((r,e,a,d,sx,sy,sz + self.INSTALL_HEIGHT))
						
						if lstate == 'V10':
							(tid,posX,posY,velX,velY,accX,accY,posZ,velZ,accZ) = struct.unpack('I9f', sbuf)
							 
							if self.dbg == True:
								print("(V10)=> (npts={:2d}: lenCnt={:4d})(idx:({:4d}) posX:{:.4f} ...accY:{:.4f} accZ:{:.4f}".format(numOfPoints,lenCount,idx,posX,accY,accZ))
						
							if (df == 'DataFrame'):
								v10df.append((self.hdr.frameNumber,'v10',tid,posX,posY,posZ + self.INSTALL_HEIGHT,velX,velY,velZ,accX,accY,accZ)) 
							else:
								v10.append((tid,posX,posY,posZ + self.INSTALL_HEIGHT ,velX,velY,velZ,accX,accY,accZ))
					
						#print("point_cloud_2d.append:[{:d}]".format(len(point_cloud_2d)))
						sbuf = b""
						
					except:
						if self.dbg == True:
							print(f"(1-8.1)Improper Type: {lstate}  Value structure found: ")
						
						return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
				
				if idx == lenCount:
					if disp == True:
						print("v1[{:d}]".format(len(v1)))
					idx = 0
					sbuf = b""
					if tlvCount <= 0: # Back to idle
						lstate = 'idle'
						if self.sm == True:
							print("(V1:{:d})=>(idle) :true".format(tlvCount))
						
						return self.list2df(True,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(True,v1,v7,v8,v9,v10,v11)
						
					else: # Go to TL to get others type value
						lstate = 'TL'
						if self.sm == True:
							print("(V1:{:d})=>(TL)".format(tlvCount))
					
				elif idx > lenCount or self.escapeCount > self.hdr.totalPackLen:
					idx = 0
					sbuf = b""
					lstate = 'idle'
					return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
				
					
			elif lstate == 'V7' : 
				sbuf += ch
				idx += 1
				self.escapeCount += 1
				if (idx%dataBytes == 0):
					
					try:
						#print(":".join("{:02x}".format(c) for c in sbuf))
						(snr,noise) = struct.unpack('2H', sbuf)
						if (df == 'DataFrame'):
							v7df.append((self.hdr.frameNumber,'v7',snr,noise))
						else:
							v7.append((snr,noise))
						
						if self.dbg == True:
							print("(V7)=> (npts={:2d}: lenCnt={:4d})(idx:({:4d}) snr:{:.4f} noise:{:.4f}".format(numOfPoints,lenCount,idx,snr,noise))
							
						sbuf = b""
					
					except:
						if self.dbg == True:
							print(f"(7.1)Improper Type {lstate} Value structure found: ")
						
						return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
					
					if tlvCount <= 0: # Back to idle
						if self.sm == True:
							print("({}):tlvCount({:d})) lenCount:{}=>(idle)".format(lstate,tlvCount,lenCount))
						lstate = 'idle'
						return self.list2df(True,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(True,v1,v7,v8,v9,v10,v11)
					
				if idx == lenCount:
					if disp == True:
						print("v7[len:{:d}]".format(len(v7)))
					idx = 0
					sbuf = b""
					if tlvCount <= 0: # Back to idle
						lstate = 'idle'
						if self.sm == True:
							print("(V7:{:d})=>(idle) :true".format(tlvCount))
						
						return self.list2df(True,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(True,v1,v7,v8,v9,v10,v11)
						
					else: # Go to TL to get others type value
						lstate = 'TL'
						if self.sm == True:
							print("(V7:{:d})=>(TL)".format(tlvCount))
							
					
				elif idx > lenCount or self.escapeCount > self.hdr.totalPackLen:
					print("(ESC)========type={:} lenth = 0 =============self.escapeCount > self.hdr.totoalPackLen==============================".format(lstate))
					idx = 0
					sbuf = b""
					lstate = 'idle'
					return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
			
			elif lstate == 'V9': #chk ok ods
				sbuf += ch
				idx += 1
				self.escapeCount += 1
				
				if (idx%dataBytes == 0):
					numOfPointsCnt += 1
					try:
						#print(":".join("{:02x}".format(c) for c in sbuf))
						(snr,noise) = struct.unpack('2H', sbuf)
						if (df == 'DataFrame'):
							v9df.append((self.hdr.frameNumber,'v9',snr,noise))
						else:
							v9.append((snr,noise))
							
						if self.dbg == True:
							print("(V9)=> (npts={:2d}: nptsCnt= {:2d} lenCnt={:4d})(idx:({:4d}) snr:{:.4f} noise:{:.4f}".format(numOfPoints,numOfPointsCnt,lenCount,idx,snr,noise))
						sbuf = b""
					
					except:
						if self.dbg == True:
							print("(9.1)Improper Type 7 Value structure found: ")
						
						return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
					
				if idx == lenCount:
					if disp == True:
						print("v9[len:{:d}]".format(len(v9)))
					idx = 0
					sbuf = b""
					if tlvCount <= 0: # Back to idle
						lstate = 'idle'
						if self.sm == True:
							print("(V9:{:d})=>(idle) :true".format(tlvCount))
						return self.list2df(True,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(True,v1,v7,v8,v9,v10,v11)
						
					else: # Go to TL to get others type value
						lstate = 'TL'
						if self.sm == True:
							print("(V9:{:d})=>(TL)".format(tlvCount))
				
				
				if idx > lenCount or self.escapeCount > self.hdr.totalPackLen:
					print("(ESC)========type={:} lenth = 0 =============self.escapeCount > self.hdr.totoalPackLen==============================".format(lstate))
					idx = 0
					sbuf = b""
					lstate = 'idle'
					return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
			
			elif lstate == 'V11':
				idx += 1
				#v11.append(ord(ch))
				
				if (df == 'DataFrame'):
					v11df.append((self.hdr.frameNumber,'v11',ord(ch)))
				else:
					v11.append(ord(ch))
					
				if idx == lenCount:
					if disp == True:
						print(f"(V11) ==> {v11}")
						print("=====V11 End====")
					sbuf = b""
					idx = 0
					lstate = 'idle'
					if self.sm == True:
						print("(V11:{:d})=>(idle)".format(tlvCount))
					
					
					return self.list2df(True,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(True,v1,v7,v8,v9,v10,v11)
				
				if idx > lenCount:
					sbuf = b""
					idx = 0
					lstate = 'idle'
					return self.list2df(False,v1df,v7df,v8df,v9df,v10df,v11df) if (df == 'DataFrame') else self.combine(False,v1,v7,v8,v9,v10,v11)
		
					
	#v17_col_names_rt =  ['fN','type','range','elv','angle','doppler','sx','sy','sz','snr','noise','Association']
	#v89_col_names_rt =  ['fN','type','X','Y','Z','doppler','snr','noise']
	def getRecordData(self,frameNum):
		s_fn =  frameNum + self.sim_startFN
		#print("frame number:{:}".format(s_fn))
		v17d = self.v17simo[(self.v17simo.fN ==  s_fn)]
		v17o = v17d.loc[:,['range','elv','angle','doppler','sx','sy','sz','snr','noise','Association','fN']].apply(pd.to_numeric)
		
		v89d = self.v89simo[(self.v89simo.fN == s_fn)]
		v89o = v89d.loc[:,['X','Y','Z','doppler','snr','noise','fN']].apply(pd.to_numeric)
		
		v10d = self.v10simo[(self.v10simo.fN == s_fn)]
		v10o = v10d.loc[:,['tid','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ','fN']].apply(pd.to_numeric)
		
		return(v17o,v89o,v10o)
		

	def readFile(self,fileName):
		#fileName = "pc32021-03-19-10-02-17.csv"  
		#df = pd.read_csv(fileName, error_bad_lines=False, warn_bad_lines=False) 
		self.fileName = fileName 
		df = pd.read_csv(self.fileName)
		
		#print("------------------- df --------------------shape:{:}".format(df.shape))
		print(df.info())
		#print(df.info(memory_usage="deep")) 
		
		self.v89simo = df[df['type'] == 'v89'].drop(['i6', 'i7', 'i8', 'i9'], axis=1).rename(columns={'i0': 'X', 'i1': 'Y', 'i2': 'Z', 'i3': 'doppler', 'i4': 'snr', 'i5': 'noise'})
		#print("======v89simo=========")
		#print(self.v89simo)
		
		self.v17simo = df[df['type'] == 'v17'].rename(columns={'i0': 'range', 'i1': 'elv', 'i2': 'angle', 'i3': 'doppler', 'i4': 'sx', 'i5': 'sy', 'i6': 'sz', 'i7': 'snr', 'i8': 'noise', 'i9': 'Association'})
		#print("======v17simo=========")
		#print(self.v17simo)
		
		#v10_col_names_rt  = ['fN','type','tid','posX','posY','posZ','velX','velY','velZ','accX','accY','accZ']
		self.v10simo = df[df['type'] == 'v10'].rename(columns={'i0': 'tid', 'i1': 'posX', 'i2': 'posY', 'i3': 'posZ', 'i4': 'velX', 'i5': 'velY', 'i6': 'velZ', 'i7': 'accX', 'i8': 'accY', 'i9': 'accZ'})
		#print("======v10simo=========")
		#print(self.v10simo)
		
		self.sim_startFN = int(df['fN'].values[1])
		self.sim_stopFN  = int(df['fN'].values[-1])
		
		#print(f'self.sim_startFN  = {self.sim_startFN} self.sim_stopFN = {self.sim_stopFN} ')
		
		return (self.sim_startFN,self.sim_stopFN,self.v17simo,self.v89simo,self.v10simo)
	
	
	def recording_ods(self,writer= None,v17=None,v89=None, v10 = None,fn = None):
		if v17 is not None:
			if len(v17) > 0:
				v17l = v17
				for i in v17l: 
					item = [fn,'v17'] + list(i)
					writer.writerow(item)
		
		if v89 is not None:
			if len(v89) > 0:
				v89l = v89
				for i in v89l: 
					item = [fn,'v89'] + list(i)
					writer.writerow(item)
		 
		if v10 is not None:
			if len(v10) > 0:
				v10l = v10
				for i in v10l: 
					item = [fn,'v10'] + list(i)
					writer.writerow(item)


